@@ -5,10 +5,8 @@ import os
 import shutil
 import unittest
 import uuid
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, timezone
 
-from dateutil import tz
-from lxml.etree import Element
 from pathlib import Path
 
 from io import BytesIO
@@ -16,7 +14,6 @@ from io import BytesIO
 from pykeepass import PyKeePass, icons
 from pykeepass.entry import Entry
 from pykeepass.group import Group
-from pykeepass.kdbx_parsing import KDBX
 from pykeepass.exceptions import BinaryError, CredentialsError, HeaderChecksumError
 
 """
@@ -33,32 +30,32 @@ Missing Tests:
   - expiry_time - get/set
 """
 
-base_dir = os.path.dirname(os.path.realpath(__file__))
+base_dir = Path(os.path.dirname(os.path.realpath(__file__)))
 logger = logging.getLogger("pykeepass")
 
 
 class KDBX3Tests(unittest.TestCase):
-    database = os.path.join(base_dir, 'test3.kdbx')
+    database = base_dir / 'test3.kdbx'
     password = 'password'
-    keyfile = os.path.join(base_dir, 'test3.key')
+    keyfile = base_dir / 'test3.key'
 
-    database_tmp = os.path.join(base_dir, 'test3_tmp.kdbx')
-    keyfile_tmp = os.path.join(base_dir, 'test3_tmp.key')
+    database_tmp = base_dir / 'test3_tmp.kdbx'
+    keyfile_tmp = base_dir / 'test3_tmp.key'
 
     # get some things ready before testing
     def setUp(self):
         shutil.copy(self.database, self.database_tmp)
         shutil.copy(self.keyfile, self.keyfile_tmp)
         self.kp = PyKeePass(
-            os.path.join(base_dir, self.database),
+            base_dir / self.database,
             password=self.password,
-            keyfile=os.path.join(base_dir, self.keyfile)
+            keyfile=base_dir / self.keyfile
         )
         # for tests which modify the database, use this
         self.kp_tmp = PyKeePass(
-            os.path.join(base_dir, self.database_tmp),
+            base_dir / self.database_tmp,
             password=self.password,
-            keyfile=os.path.join(base_dir, self.keyfile_tmp)
+            keyfile=base_dir / self.keyfile_tmp
         )
 
     def tearDown(self):
@@ -67,12 +64,12 @@ class KDBX3Tests(unittest.TestCase):
 
 
 class KDBX4Tests(KDBX3Tests):
-    database = os.path.join(base_dir, 'test4.kdbx')
+    database = base_dir / 'test4.kdbx'
     password = 'password'
-    keyfile = os.path.join(base_dir, 'test4.key')
+    keyfile = base_dir / 'test4.key'
 
-    database_tmp = os.path.join(base_dir, 'test4_tmp.kdbx')
-    keyfile_tmp = os.path.join(base_dir, 'test4_tmp.key')
+    database_tmp = base_dir / 'test4_tmp.kdbx'
+    keyfile_tmp = base_dir / 'test4_tmp.key'
 
 
 class EntryFindTests3(KDBX3Tests):
@@ -144,14 +141,23 @@ class EntryFindTests3(KDBX3Tests):
         self.assertEqual(len(results), 1)
         self.assertEqual(results[0].autotype_sequence, '{USERNAME}{TAB}{PASSWORD}{ENTER}')
 
+    def test_find_entries_by_autotype_window(self):
+        results = self.kp.find_entries(autotype_window='test', regex=True, flags="i")
+        self.assertEqual(len(results), 1)
+        self.assertEqual(results[0].autotype_window, 'TEST')
+
     def test_find_entries_by_autotype_enabled(self):
         results = self.kp.find_entries(autotype_enabled=True)
         self.assertEqual(len(results), len(self.kp.entries) - 1)
 
     def test_find_entries_by_otp(self):
-        results = self.kp.find_entries(otp='otpsecret', regex=True, flags='i')
+        results = self.kp.find_entries(otp='nonmatch', regex=True, flags='i')
+        self.assertEqual(len(results), 0)
+        results = self.kp.find_entries(otp='OTPSECRETT', regex=True)
         self.assertEqual(len(results), 1)
         self.assertEqual('foobar_entry', results[0].title)
+        import pyotp
+        self.assertEqual(len(pyotp.parse_uri(results[0].otp).now()), 6)
 
     def test_find_entries(self):
         results = self.kp.find_entries(title='Root_entry', regex=True)
@@ -219,7 +225,7 @@ class EntryFindTests3(KDBX3Tests):
 
     def test_add_delete_move_entry(self):
         unique_str = 'test_add_entry_'
-        expiry_time = datetime.now()
+        expiry_time = datetime.now(timezone.utc)
         entry = self.kp.add_entry(
             self.kp.root_group,
             unique_str + 'title',
@@ -243,8 +249,6 @@ class EntryFindTests3(KDBX3Tests):
         self.assertEqual(len(results.tags), 6)
         self.assertTrue(results.uuid != None)
         self.assertTrue(results.autotype_sequence is None)
-        # convert naive datetime to utc
-        expiry_time_utc = expiry_time.replace(tzinfo=tz.gettz()).astimezone(tz.gettz('UTC'))
         self.assertEqual(results.icon, icons.KEY)
 
         sub_group = self.kp.add_group(self.kp.root_group, 'sub_group')
@@ -280,6 +284,52 @@ class EntryFindTests3(KDBX3Tests):
         )
         self.assertRaises(Exception, entry)
 
+    # ---------- Timezone test -----------
+
+    def test_expiration_time_tz(self):
+        # The expiration date is compared in UTC
+        # setting expiration date with tz offset 6 hours should result in expired entry
+        unique_str = 'test_exptime_tz_1_'
+        expiry_time = datetime.now(timezone(offset=timedelta(hours=6))).replace(microsecond=0)
+        self.kp.add_entry(
+            self.kp.root_group,
+            unique_str + 'title',
+            unique_str + 'user',
+            unique_str + 'pass',
+            expiry_time=expiry_time
+        )
+        results = self.kp.find_entries_by_title(unique_str + 'title', first=True)
+        self.assertEqual(results.expired, True)
+        self.assertEqual(results.expiry_time, expiry_time.astimezone(timezone.utc))
+
+        # setting expiration date with UTC tz should result in expired entry
+        unique_str = 'test_exptime_tz_2_'
+        expiry_time = datetime.now(timezone.utc).replace(microsecond=0)
+        self.kp.add_entry(
+            self.kp.root_group,
+            unique_str + 'title',
+            unique_str + 'user',
+            unique_str + 'pass',
+            expiry_time=expiry_time
+        )
+        results = self.kp.find_entries_by_title(unique_str + 'title', first=True)
+        self.assertEqual(results.expired, True)
+        self.assertEqual(results.expiry_time, expiry_time.astimezone(timezone.utc))
+
+        # setting expiration date with tz offset -6 hours while adding 6 hours should result in valid entry
+        unique_str = 'test_exptime_tz_3_'
+        expiry_time = datetime.now(timezone(offset=timedelta(hours=-6))).replace(microsecond=0) + timedelta(hours=6)
+        self.kp.add_entry(
+            self.kp.root_group,
+            unique_str + 'title',
+            unique_str + 'user',
+            unique_str + 'pass',
+            expiry_time=expiry_time
+        )
+        results = self.kp.find_entries_by_title(unique_str + 'title', first=True)
+        self.assertEqual(results.expired, False)
+        self.assertEqual(results.expiry_time, expiry_time.astimezone(timezone.utc))
+
     # ---------- Entries representation -----------
 
     def test_print_entries(self):
@@ -430,7 +480,7 @@ class RecycleBinTests3(KDBX3Tests):
 class EntryTests3(KDBX3Tests):
 
     def test_fields(self):
-        time = datetime.now().replace(microsecond=0)
+        expiry_time = datetime.now(timezone.utc).replace(microsecond=0)
         entry = Entry(
             'title',
             'username',
@@ -440,7 +490,7 @@ class EntryTests3(KDBX3Tests):
             tags='tags',
             otp='otp',
             expires=True,
-            expiry_time=time,
+            expiry_time=expiry_time,
             icon=icons.KEY,
             kp=self.kp
         )
@@ -453,8 +503,7 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(entry.tags, ['tags'])
         self.assertEqual(entry.otp, 'otp')
         self.assertEqual(entry.expires, True)
-        self.assertEqual(entry.expiry_time,
-                         time.replace(tzinfo=tz.gettz()).astimezone(tz.gettz('UTC')))
+        self.assertEqual(entry.expiry_time, expiry_time)
         self.assertEqual(entry.icon, icons.KEY)
         self.assertEqual(entry.is_a_history_entry, False)
         self.assertEqual(
@@ -484,7 +533,7 @@ class EntryTests3(KDBX3Tests):
         self.assertNotEqual(clone1, clone2)
 
     def test_set_and_get_fields(self):
-        time = datetime.now().replace(microsecond=0)
+        time = datetime.now(timezone.utc).replace(microsecond=0)
         changed_time = time + timedelta(hours=9)
         changed_string = 'changed_'
         entry = Entry(
@@ -525,8 +574,7 @@ class EntryTests3(KDBX3Tests):
         self.assertEqual(entry.get_custom_property('foo'), None)
         # test time properties
         self.assertEqual(entry.expires, False)
-        self.assertEqual(entry.expiry_time,
-                         changed_time.replace(tzinfo=tz.gettz()).astimezone(tz.gettz('UTC')))
+        self.assertEqual(entry.expiry_time, changed_time)
 
         entry.tags = 'changed_tags'
         self.assertEqual(entry.tags, ['changed_tags'])
@@ -537,8 +585,8 @@ class EntryTests3(KDBX3Tests):
 
     def test_expired_datetime_offset(self):
         """Test for https://github.com/pschmitt/pykeepass/issues/115"""
-        future_time = datetime.now() + timedelta(days=1)
-        past_time = datetime.now() - timedelta(days=1)
+        future_time = datetime.now(timezone.utc) + timedelta(days=1)
+        past_time = datetime.now(timezone.utc) - timedelta(days=1)
         entry = Entry(
             'title',
             'username',
@@ -611,6 +659,17 @@ class EntryTests3(KDBX3Tests):
         self.assertFalse(e.is_custom_property_protected('not-protected'))
         self.assertFalse(e.is_custom_property_protected('non-existent'))
 
+    def test_reindex(self):
+        e1 = self.kp.add_entry(self.kp.root_group, 'Test-Index1', 'user-index', 'pass')
+        e2 = self.kp.add_entry(self.kp.root_group, 'Test-Index2', 'user-index', 'pass')
+        e3 = self.kp.add_entry(self.kp.root_group, 'Test-Index3', 'user-index', 'pass')
+        e4 = self.kp.add_entry(self.kp.root_group, 'Test-Index4', 'user-index', 'pass')
+        e2.reindex(0)
+        e3.reindex(0)
+        e4.reindex(0)
+        entries = self.kp.find_entries(username="user-index")
+        self.assertEqual(entries, [e4,e3,e2,e1])
+
 
 class EntryHistoryTests3(KDBX3Tests):
 
@@ -681,7 +740,7 @@ class EntryHistoryTests3(KDBX3Tests):
 
         # change the active entries to test integrity of the history items
         backup = {}
-        now = datetime.now()
+        now = datetime.now(timezone.utc)
         for entry in res1:
             backup[entry.uuid] = {"atime": entry.atime, "mtime": entry.mtime, "ctime": entry.ctime}
             entry.title = changed + 'title'
@@ -823,7 +882,7 @@ class PyKeePassTests3(KDBX3Tests):
 
     def test_set_credentials(self):
         self.kp_tmp.password = 'f00bar'
-        self.kp_tmp.keyfile = os.path.join(base_dir, 'change.key')
+        self.kp_tmp.keyfile = base_dir / 'change.key'
         self.kp_tmp.save()
         self.kp_tmp = PyKeePass(
             self.kp_tmp.filename,
@@ -849,8 +908,8 @@ class PyKeePassTests3(KDBX3Tests):
 
         required_days = 5
         recommended_days = 5
-        unexpired_date = datetime.now() - timedelta(days=1)
-        expired_date = datetime.now() - timedelta(days=10)
+        unexpired_date = datetime.now(timezone.utc) - timedelta(days=1)
+        expired_date = datetime.now(timezone.utc) - timedelta(days=10)
 
         self.kp.credchange_required_days = required_days
         self.kp.credchange_recommended_days = recommended_days
@@ -960,7 +1019,8 @@ class BugRegressionTests3(KDBX3Tests):
     def test_issue344(self):
         # accessing expiry_time throws exception when None
 
-        e = self.kp.find_entries(title='none_date', first=True)
+        e = self.kp_tmp.find_entries(title='none_date', first=True)
+        e._element.xpath('Times/ExpiryTime')[0].text = None
         self.assertEqual(e.expiry_time, None)
 
 class EntryFindTests4(KDBX4Tests, EntryFindTests3):
@@ -990,7 +1050,7 @@ class BugRegressionTests4(KDBX4Tests, BugRegressionTests3):
 
 class CtxManagerTests(unittest.TestCase):
     def test_ctx_manager(self):
-        with PyKeePass(os.path.join(base_dir, 'test4.kdbx'), password='password', keyfile=base_dir + '/test4.key') as kp:
+        with PyKeePass(base_dir / 'test4.kdbx', password='password', keyfile=base_dir / 'test4.key') as kp:
             results = kp.find_entries_by_username('foobar_user', first=True)
             self.assertEqual('foobar_user', results.username)
 
@@ -1000,40 +1060,43 @@ class KDBXTests(unittest.TestCase):
     def test_open_save(self):
         """try to open all databases, save them, then open the result"""
 
-        with open(os.path.join(base_dir, 'test3.kdbx'), 'rb') as file:
+        # for database stream open test
+        with open(base_dir / 'test3.kdbx', 'rb') as file:
             stream = BytesIO(file.read())
+        # for keyfile file descriptor test
+        keyfile_fd = open(base_dir / 'test4.key', 'rb')
 
         filenames_in = [
-            os.path.join(base_dir, 'test3.kdbx'),                 # KDBX v3
-            Path(base_dir).joinpath('test4.kdbx'),                # KDBX v4 (and test pathlib)
-            os.path.join(base_dir, 'test4_aes.kdbx'),             # KDBX v4 AES
-            os.path.join(base_dir, 'test4_aeskdf.kdbx'),          # KDBX v3 AESKDF
-            os.path.join(base_dir, 'test4_chacha20.kdbx'),        # KDBX v4 ChaCha
-            os.path.join(base_dir, 'test4_twofish.kdbx'),         # KDBX v4 Twofish
-            os.path.join(base_dir, 'test4_hex.kdbx'),             # legacy 64 byte hexadecimal keyfile
-            os.path.join(base_dir, 'test3_transformed.kdbx'),     # KDBX v3 transformed_key open
-            os.path.join(base_dir, 'test4_transformed.kdbx'),     # KDBX v4 transformed_key open
+            base_dir / 'test3.kdbx',                 # KDBX v3
+            base_dir / 'test4_aes.kdbx',             # KDBX v4 AES
+            base_dir / 'test4_aeskdf.kdbx',          # KDBX v3 AESKDF
+            base_dir / 'test4_chacha20.kdbx',        # KDBX v4 ChaCha
+            base_dir / 'test4_twofish.kdbx',         # KDBX v4 Twofish
+            base_dir / 'test4_hex.kdbx',             # legacy 64 byte hexadecimal keyfile
+            base_dir / 'test3_transformed.kdbx',     # KDBX v3 transformed_key open
+            base_dir / 'test4_transformed.kdbx',     # KDBX v4 transformed_key open
             stream,                                               # test stream opening
-            os.path.join(base_dir, 'test4_aes_uncompressed.kdbx'),# KDBX v4 AES uncompressed
-            os.path.join(base_dir, 'test4_twofish_uncompressed.kdbx'),# KDBX v4 Twofish uncompressed
-            os.path.join(base_dir, 'test4_chacha20_uncompressed.kdbx'),# KDBX v4 ChaCha uncompressed
-            os.path.join(base_dir, 'test4_argon2id.kdbx'),        # KDBX v4 Argon2id
+            base_dir / 'test4_aes_uncompressed.kdbx',# KDBX v4 AES uncompressed
+            base_dir / 'test4_twofish_uncompressed.kdbx',# KDBX v4 Twofish uncompressed
+            base_dir / 'test4_chacha20_uncompressed.kdbx',# KDBX v4 ChaCha uncompressed
+            base_dir / 'test4_argon2id.kdbx',        # KDBX v4 Argon2id
+            base_dir / 'test4.kdbx',        # KDBX v4 with keyfile file descriptor
         ]
         filenames_out = [
-            os.path.join(base_dir, 'test3.kdbx.out'),
-            Path(base_dir).joinpath('test4.kdbx.out'),
-            os.path.join(base_dir, 'test4_aes.kdbx.out'),
-            os.path.join(base_dir, 'test4_aeskdf.kdbx.out'),
-            os.path.join(base_dir, 'test4_chacha20.kdbx.out'),
-            os.path.join(base_dir, 'test4_twofish.kdbx.out'),
-            os.path.join(base_dir, 'test4_hex.kdbx.out'),
-            os.path.join(base_dir, 'test3_transformed.kdbx.out'),
-            os.path.join(base_dir, 'test4_transformed.kdbx.out'),
+            base_dir / 'test3.kdbx.out',
+            base_dir / 'test4_aes.kdbx.out',
+            base_dir / 'test4_aeskdf.kdbx.out',
+            base_dir / 'test4_chacha20.kdbx.out',
+            base_dir / 'test4_twofish.kdbx.out',
+            base_dir / 'test4_hex.kdbx.out',
+            base_dir / 'test3_transformed.kdbx.out',
+            base_dir / 'test4_transformed.kdbx.out',
             BytesIO(),
-            os.path.join(base_dir, 'test4_aes_uncompressed.kdbx.out'),
-            os.path.join(base_dir, 'test4_twofish_uncompressed.kdbx.out'),# KDBX v4 Twofish uncompressed
-            os.path.join(base_dir, 'test4_chacha20_uncompressed.kdbx.out'),# KDBX v4 ChaCha uncompressed
-            os.path.join(base_dir, 'test4_argon2id.kdbx.out'),
+            base_dir / 'test4_aes_uncompressed.kdbx.out',
+            base_dir / 'test4_twofish_uncompressed.kdbx.out',# KDBX v4 Twofish uncompressed
+            base_dir / 'test4_chacha20_uncompressed.kdbx.out',# KDBX v4 ChaCha uncompressed
+            base_dir / 'test4_argon2id.kdbx.out',
+            base_dir / 'test4.kdbx.out',        # KDBX v4 with keyfile file descriptor
         ]
         passwords = [
             'password',
@@ -1042,9 +1105,9 @@ class KDBXTests(unittest.TestCase):
             'password',
             'password',
             'password',
+            None,
+            None,
             'password',
-            None,
-            None,
             'password',
             'password',
             'password',
@@ -1058,7 +1121,6 @@ class KDBXTests(unittest.TestCase):
             None,
             None,
             None,
-            None,
             b'\xfb\xb1!\x0e0\x94\xd4\x868\xa5\x04\xe6T\x9b<\xf9+\xb8\x82EN\xbc\xbe\xbc\xc8\xd3\xbbf\xfb\xde\xff.',
             b'\x95\x0be\x9ca\x9e<\xe0\x07\x02\x7f\xc3\xd8\xa1\xa6&\x985\x8f!\xa6\x18k\x13\xa2\xd2\r=\xf3\xebd\xc5',
             None,
@@ -1066,26 +1128,26 @@ class KDBXTests(unittest.TestCase):
             None,
             None,
             None,
-        ]
+            None,
+ ]
         keyfiles = [
-            'test3.key',
-            Path('test4.key'),
-            'test4.key',
-            'test4.key',
-            'test4.key',
-            'test4.key',
-            'test4_hex.key',
+            base_dir / 'test3.key',
+            base_dir / 'test4.key',
+            base_dir / 'test4.key',
+            base_dir / 'test4.key',
+            base_dir / 'test4.key',
+            base_dir / 'test4_hex.key',
             None,
             None,
-            'test3.key',
+            base_dir / 'test3.key',
             None,
             None,
             None,
             None,
+            keyfile_fd
         ]
         encryption_algorithms = [
             'aes256',
-            'chacha20',
             'aes256',
             'aes256',
             'chacha20',
@@ -1098,11 +1160,11 @@ class KDBXTests(unittest.TestCase):
             'twofish',
             'chacha20',
             'aes256',
+            'chacha20',
         ]
         kdf_algorithms = [
             'aeskdf',
             'argon2',
-            'argon2',
             'aeskdf',
             'argon2',
             'argon2',
@@ -1114,6 +1176,7 @@ class KDBXTests(unittest.TestCase):
             'argon2',
             'argon2',
             'argon2id',
+            'argon2',
         ]
         versions = [
             (3, 1),
@@ -1122,7 +1185,6 @@ class KDBXTests(unittest.TestCase):
             (4, 0),
             (4, 0),
             (4, 0),
-            (4, 0),
             (3, 1),
             (4, 0),
             (3, 1),
@@ -1130,6 +1192,7 @@ class KDBXTests(unittest.TestCase):
             (4, 0),
             (4, 0),
             (4, 0),
+            (4, 0),
         ]
 
         for (filename_in, filename_out, password, transformed_key,
@@ -1140,7 +1203,7 @@ class KDBXTests(unittest.TestCase):
             kp = PyKeePass(
                 filename_in,
                 password,
-                None if keyfile is None else os.path.join(base_dir, keyfile),
+                keyfile,
                 transformed_key=transformed_key
             )
             self.assertEqual(kp.encryption_algorithm, encryption_algorithm)
@@ -1159,13 +1222,14 @@ class KDBXTests(unittest.TestCase):
             kp = PyKeePass(
                 filename_out,
                 password,
-                None if keyfile is None else os.path.join(base_dir, keyfile),
+                keyfile,
                 transformed_key=transformed_key
             )
 
-        for filename in os.listdir(base_dir):
-            if filename.endswith('.out'):
-                os.remove(os.path.join(base_dir, filename))
+        for filename in base_dir.glob('*.out'):
+            os.remove(filename)
+
+        keyfile_fd.close()
 
 
     def test_open_error(self):
@@ -1175,6 +1239,8 @@ class KDBXTests(unittest.TestCase):
             'test3.kdbx',
             'test4.kdbx',
             'test4.kdbx',
+            'test4.kdbx',
+            'test4.kdbx',
             'test3.key',
         ]
         passwords = [
@@ -1183,12 +1249,15 @@ class KDBXTests(unittest.TestCase):
             'invalid',
             'password',
             'password',
+            'password',
+            'password',
         ]
         keyfiles = [
             'test3.key',
             'test4.key',
             'test4.key',
-            'test3.key',
+            'test_invalidversion.key',
+            'test.svg',
             'test3.key',
         ]
         errors = [
@@ -1196,16 +1265,49 @@ class KDBXTests(unittest.TestCase):
             CredentialsError,
             CredentialsError,
             CredentialsError,
+            CredentialsError,
+            CredentialsError,
             HeaderChecksumError,
         ]
         for database, password, keyfile, error in zip(databases, passwords, keyfiles, errors):
             with self.assertRaises(error):
                 PyKeePass(
-                    os.path.join(base_dir, database),
+                    base_dir / database,
                     password,
-                    os.path.join(base_dir, keyfile)
+                    base_dir / keyfile
                 )
 
+
+    def test_open_no_decrypt(self):
+        """Open database but do not decrypt payload.  Needed for reading header data for OTP tokens"""
+
+
+        databases = [
+            'test3.kdbx',
+            'test4.kdbx',
+        ]
+        passwords = [
+            'invalid_password',
+            'invalid_password',
+        ]
+        enc_algs = [
+            'aes256',
+            'chacha20'
+        ]
+        versions = [
+            (3, 1),
+            (4, 0),
+        ]
+        for database, password, enc_alg, version in zip(databases, passwords, enc_algs, versions):
+            kp = PyKeePass(
+                os.path.join(base_dir, database),
+                password,
+                decrypt=False
+            )
+
+            self.assertEqual(kp.encryption_algorithm, enc_alg)
+            self.assertEqual(kp.version, version)
+
 if __name__ == '__main__':
     unittest.main()
 
